local idTweaks = LibStub('AceAddon-3.0'):GetAddon('idTweaks')
local ActionBarsModule = idTweaks:NewModule('ActionBars', 'AceConsole-3.0', 'AceEvent-3.0')
local LibActionButton = LibStub('LibActionButton-1.0')

local TL, TC, TR = 'TOPLEFT', 'TOP', 'TOPRIGHT'
local ML, MC, MR = 'LEFT', 'CENTER', 'RIGHT'
local BL, BC, BR = 'BOTTOMLEFT', 'BOTTOM', 'BOTTOMRIGHT'

ActionBarsModule.DEBUG = false

function ActionBarsModule:Debug(...)
  if ActionBarsModule.DEBUG then
    self:Print(...)
  end
end

function ActionBarsModule:setDefaults()
  self:Debug('ActionBarsModule:setDefaults')

  self.defaults = {
    profile = {
      enabled = true,
      barPadding = 3,
      barMargin = 3,
      buttonPadding = 3,
    }
  }
end

function ActionBarsModule:setDatabase()
  self:Debug('ActionBarsModule:setDatabase')

  self.db = idTweaks.db:RegisterNamespace('ActionBars', self.defaults)
end

function ActionBarsModule:setOptions()
  self:Debug('ActionBarsModule:setOptions')

  self.options = {
    name = 'ActionBars',
    handler = self,
    type = 'group',
    args = {
      enabled = {
        order = 1,
        name = 'Enabled',
        desc = 'WARNING: Reloads the UI!',
        handler = self,
        type = 'toggle',
        get = function()
          return self.db.profile.enabled
        end,
        set = function(_, value)
          local MinimapModule = idTweaks:GetModule('Minimap')
          self.db.profile.enabled = value
          self:check()
          MinimapModule.shape.position:check()

          ReloadUI()
        end,
      },

      size = {
        name = 'Size',
        desc = 'Control the size of the frame.',
        type = 'group',
        handler = self,
        args = {
          barMargin = {
            name = 'Bar Margin',
            type = 'range',
            min = 0,
            max = 50,
            step = 1,
            bigStep = 1,
            get = function()
              return self.db.profile.barMargin
            end,
            set = function(_, value)
              self.db.profile.barMargin = value
              self:sizeAllBars()
              self:positionAllBars()
            end
          },

          barPadding = {
            name = 'Bar Padding',
            type = 'range',
            min = 0,
            max = 50,
            step = 1,
            bigStep = 1,
            get = function()
              return self.db.profile.barPadding
            end,
            set = function(_, value)
              self.db.profile.barPadding = value
              self:sizeAllBars()
              self:positionAllBars()
            end
          },

          buttonPadding = {
            name = 'Button Padding',
            type = 'range',
            min = 0,
            max = 50,
            step = 1,
            bigStep = 1,
            get = function()
              return self.db.profile.buttonPadding
            end,
            set = function(_, value)
              self.db.profile.buttonPadding = value
              self:sizeAllBars()
              self:positionAllBars()
            end
          },
        },
      },
    },
  }
end

function ActionBarsModule:registerOptions()
  self:Debug('ActionBarsModule:registerOptions')

  LibStub('AceConfig-3.0'):RegisterOptionsTable('ActionBars', self.options)
  LibStub('AceConfigDialog-3.0'):AddToBlizOptions('ActionBars', 'ActionBars', 'idTweaks')
end

function ActionBarsModule:OnInitialize()
  self:Debug('ActionBarsModule:OnInitialize')

  self.frames = {}

  self:setDefaults()
  self:setDatabase()
  self:setOptions()
end

function ActionBarsModule:OnEnable()
  self:Debug('ActionBarsModule:OnEnable')

  self:registerOptions()

  self:check()
end

function ActionBarsModule:OnDisable()
  self:Debug('ActionBarsModule:OnDisable')

  self:disable()
end

function ActionBarsModule:check()
  self:Debug('ActionBarsModule:check')

  if self.db.profile.enabled then
    self:enable()
  else
    self:disable()
  end
end

function ActionBarsModule:enable()
  self:Debug('ActionBarsModule:enable')

  self:disableBlizzardBars()
  self:enableActionBars()
  self:enablePetBar()

  self:sizeAllBars()
  self:positionAllBars()
end

function ActionBarsModule:disable()
  self:Debug('ActionBarsModule:disable')

  self:disablePetBar()
  self:disableActionBars()
  self:enableBlizzardBars()
end

function ActionBarsModule:enableActionBars()
  self:Debug('ActionBarsModule:enableActionBars')

  self.frames.bar1 = self.frames.bar1 or self:createBar('idTweaksActionBar1', 'action', 1, 12)
  self.frames.bar2 = self.frames.bar2 or self:createBar('idTweaksActionBar2', 'action', 13, 24)
  self.frames.bar3 = self.frames.bar3 or self:createBar('idTweaksActionBar3', 'action', 25, 36)
  self.frames.bar4 = self.frames.bar4 or self:createBar('idTweaksActionBar4', 'action', 37, 48)
  self.frames.bar5 = self.frames.bar5 or self:createBar('idTweaksActionBar5', 'action', 49, 60)
end

function ActionBarsModule:disableActionBars()
  self:Debug('ActionBarsModule:disableActionBars')

  if self.frames.bar1 then
    self.frames.bar1:ClearAllPoints()
    self.frames.bar1:SetPoint(BR, UIParent, TL, 0, 0)
  end
end

function ActionBarsModule:enablePetBar()
  self:Debug('ActionBarsModule:enablePetBar')

  self.frames.petbar = self.frames.petbar or self:createPetBar('idTweaksActionBarPet', 'action', 1, 10)
  self.frames.petbar:Hide()

  local function petEvent(event)
    self:Debug('ActionBarsModule:enablePetBar petEvent', event)

    if InCombatLockdown() then return end

    if PetHasActionBar() then
      if not self.frames.petbar:IsShown() then
        self.frames.petbar:Show()
      end
    else
      self.frames.petbar:Hide()
    end
  end

  self:RegisterEvent('PLAYER_CONTROL_LOST', petEvent)
  self:RegisterEvent('PLAYER_CONTROL_GAINED', petEvent)
  self:RegisterEvent('PLAYER_FARSIGHT_FOCUS_CHANGED', petEvent)
  self:RegisterEvent('UNIT_PET', petEvent)
  self:RegisterEvent('UNIT_FLAGS', petEvent)
  self:RegisterEvent('PET_BAR_UPDATE', petEvent)
  self:RegisterEvent('PET_BAR_UPDATE_COOLDOWN', petEvent)
  self:RegisterEvent('PET_BAR_SHOWGRID', petEvent)
  self:RegisterEvent('PET_BAR_HIDEGRID', petEvent)
  self:RegisterEvent('PET_BAR_UPDATE_USABLE', petEvent)
  self:RegisterEvent('PET_UI_UPDATE', petEvent)
  self:RegisterEvent('PLAYER_TARGET_CHANGED', petEvent)
  self:RegisterEvent('PLAYER_MOUNT_DISPLAY_CHANGED', petEvent)
end

function ActionBarsModule:disablePetBar()
  self:Debug('ActionBarsModule:disablePetBar')

  if not self.frames.petbar then
    self:Debug('ActionBarsModule:disablePetBar there is no pet bar')
    return
  end

  self:UnregisterEvent('PLAYER_CONTROL_LOST')
  self:UnregisterEvent('PLAYER_CONTROL_GAINED')
  self:UnregisterEvent('PLAYER_FARSIGHT_FOCUS_CHANGED')
  self:UnregisterEvent('UNIT_PET')
  self:UnregisterEvent('UNIT_FLAGS')
  self:UnregisterEvent('PET_BAR_UPDATE')
  self:UnregisterEvent('PET_BAR_UPDATE_COOLDOWN')
  self:UnregisterEvent('PET_BAR_SHOWGRID')
  self:UnregisterEvent('PET_BAR_HIDEGRID')
  self:UnregisterEvent('PET_BAR_UPDATE_USABLE')
  self:UnregisterEvent('PET_UI_UPDATE')
  self:UnregisterEvent('PLAYER_TARGET_CHANGED')
  self:UnregisterEvent('PLAYER_MOUNT_DISPLAY_CHANGED')

  PetActionBarFrame:EnableMouse('true')
  SlidingActionBarTexture0:SetAlpha(1)
  SlidingActionBarTexture1:SetAlpha(1)

  for i = self.frames.petbar.firstId, self.frames.petbar.lastId do
    local button = _G['PetActionButton' .. i]
    button:ClearAllPoints()
    button:SetParent(PetActionBarFrame)
    button:SetPoint(BR, PetActionBarFrame, BL, 65 + ((i - 1) * 38), 0)
  end

  self.frames.petbar:Hide()
end

function ActionBarsModule:enableBlizzardBars()
  self:Debug('ActionBarsModule:enableBlizzardBars')

  MainMenuBar:SetScript('OnShow', nil)
  MainMenuBar:Show()

  MultiBarLeft:SetScript('OnShow', nil)
  MultiBarLeft:Show()

  MultiBarRight:SetScript('OnShow', nil)
  MultiBarRight:Show()
end

function ActionBarsModule:disableBlizzardBars()
  self:Debug('ActionBarsModule:disableBlizzardBars')

  MainMenuBar:SetScript('OnShow', MainMenuBar.Hide)
  MainMenuBar:Hide()

  MultiBarLeft:SetScript('OnShow', MultiBarLeft.Hide)
  MultiBarLeft:Hide()

  MultiBarRight:SetScript('OnShow', MultiBarRight.Hide)
  MultiBarRight:Hide()
end

function ActionBarsModule:createButton(i, buttonName, frameHeader)
  self:Debug('ActionBarsModule:createButton', i, buttonName)

  local button = LibActionButton:CreateButton(i, buttonName, frameHeader, {
    outOfRangeColoring = 'button',
    tooltip = 'enabled',
    showGrid = false,
    colors = {
      range = { 0.8, 0.1, 0.1 },
      mana = { 0.5, 0.5, 1.0 }
    },
    hideElements = {
      macro = false,
      hotkey = false,
      equipped = false,
    },
    keyBoundTarget = false,
    clickOnDown = false,
    flyoutDirection = 'UP',
  })

  local icon = _G[buttonName .. 'Icon']
  local texture = (_G[buttonName .. 'NormalTexture2'] or _G[buttonName .. 'NormalTexture'])
  local hotkey = _G[buttonName .. 'HotKey']
  local macrotext = _G[buttonName .. 'Name']

  icon:SetTexCoord(0.08,0.92,0.08,0.92)

  hotkey:Hide()
  macrotext:Hide()

  texture:SetAlpha(0.5)
  texture:SetTexCoord(0,0,0,0)

  -- ugly but works
  texture.SetAlpha = nothing
  hotkey.Show = nothing
  macrotext.Show = nothing

  return button
end

function ActionBarsModule:createBar(barName, barType, firstId, lastId)
  self:Debug('ActionBarsModule:createBar', barName, barType, firstId, lastId)

  local frameHeader = CreateFrame('Frame', barName .. 'Header', UIParent, 'SecureHandlerStateTemplate')
  local bar = CreateFrame('Frame', barName, UIParent)

  bar.header = frameHeader
  bar.barName = barName
  bar.barType = barType
  bar.buttonName = barName .. 'Button'
  bar.firstId = firstId
  bar.lastId = lastId

  bar.background = bar:CreateTexture(barName .. 'Background', 'BACKGROUND')
  bar.background:SetAllPoints(bar)
  bar.background:SetColorTexture(0, 0, 0, 0.5)

  local j = 1
  for i = bar.firstId, bar.lastId do
    bar['button' .. j] = self:createButton(i, bar.buttonName .. j, bar.header)
    j = j + 1
  end

  self:sizeBar(bar)
  self:setBarActions(bar, bar.barType, firstId, lastId)

  RegisterStateDriver(bar.header, 'page', '0')
  bar.header:SetAttribute('_onstate-page', [[
    self:SetAttribute('state', newstate)
    control:ChildUpdate('state', newstate)
  ]])

  return bar
end

function ActionBarsModule:createPetBar(barName, barType, firstId, lastId)
  self:Debug('ActionBarsModule:createPetBar', barName, barType, firstId, lastId)

  local frameHeader = CreateFrame('Frame', barName .. 'Header', UIParent, 'SecureHandlerStateTemplate')
  local bar = CreateFrame('Frame', barName, UIParent)

  bar.header = frameHeader
  bar.barName = barName
  bar.barType = barType
  bar.buttonName = barName .. 'Button'
  bar.firstId = firstId
  bar.lastId = lastId

  bar.background = bar:CreateTexture(barName .. 'Background', 'BACKGROUND')
  bar.background:SetAllPoints(bar)
  bar.background:SetColorTexture(0, 0, 0, 0.5)

  local j = 1
  for i = bar.firstId, bar.lastId do
    local button = _G['PetActionButton' .. i]
    local buttonIcon = _G['PetActionButton' .. i .. 'Icon']
    local buttonBorder = _G['PetActionButton' .. i .. 'NormalTexture2']

    button:ClearAllPoints()
    button:SetParent(bar)
    button:SetWidth(idTweaksActionBar1Button1:GetWidth())
    button:SetHeight(idTweaksActionBar1Button1:GetHeight())

    buttonIcon:SetTexCoord(0.08,0.92,0.08,0.92)
    buttonBorder:Hide()

    if button:HasScript('OnUpdate') then
      button:HookScript('OnUpdate', function()
        if self.frames.petbar then
          buttonBorder:Hide()
        end
      end)
    end

    bar['button' .. j] = button

    j = j + 1
  end

  PetActionBarFrame:EnableMouse('false')
  SlidingActionBarTexture0:SetAlpha(0)
  SlidingActionBarTexture1:SetAlpha(0)

  self:sizeBar(bar)

  return bar
end

function ActionBarsModule:setBarActions(bar, barType, firstId, lastId)
  self:Debug('ActionBarsModule:setBarActions', barType, firstId, lastId)

  for i = 1, lastId - firstId + 1 do
    bar['button'..i]:SetState(0, barType, firstId - 1 + i)
  end
end

function ActionBarsModule:sizeBar(bar)
  self:Debug('ActionBarsModule:sizeBar', bar.barName)

  local barPadding = self.db.profile.barPadding
  local buttonPadding = self.db.profile.buttonPadding

  local j = 1
  for i = bar.firstId, bar.lastId do
    local button = bar['button' .. j]

    if j == 1 then
      button:SetPoint(ML, bar, ML, barPadding, 0)
    else
      button:SetPoint(ML, bar['button' .. j - 1], MR, buttonPadding, 0)
    end

    j = j + 1
  end

  local buttonAmount = bar.lastId + 1 - bar.firstId
  local buttonWidth = bar.button1:GetWidth()
  local buttonHeight = bar.button1:GetHeight()

  local width = buttonWidth * buttonAmount + buttonPadding * (buttonAmount - 1) + barPadding * 2
  local height = buttonHeight + barPadding * 2

  bar:SetWidth(width)
  bar:SetHeight(height)
end

function ActionBarsModule:sizeAllBars()
  self:Debug('ActionBarsModule:sizeAllBars')

  for b = 1, 5 do
    local bar = self.frames['bar'..b]

    self:sizeBar(bar)
  end
end

function ActionBarsModule:positionAllBars()
  self:Debug('ActionBarsModule:positionAllBars')

  local barMargin = ActionBarsModule.db.profile.barMargin

  self.frames.bar1:ClearAllPoints()
  self.frames.bar1:SetPoint(BC, UIParent, BC, 0, barMargin)
  self.frames.bar2:SetPoint(BC, self.frames.bar1, TC, 0, barMargin)
  self.frames.bar3:SetPoint(BC, self.frames.bar2, TC, 0, barMargin)
  self.frames.bar4:SetPoint(BC, self.frames.bar3, TC, 0, barMargin)
  self.frames.bar5:SetPoint(BC, self.frames.bar4, TC, 0, barMargin)
  self.frames.petbar:SetPoint(BC, self.frames.bar5, TC, 0, barMargin)
end
